# Copyright © 2014  Émilien Tlapale
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import os
import os.path
import pathlib
import xml.dom.minidom

import citeproc as cp
import citeproc.formatter

import bibgen.formatter

docbook_ns = 'http://docbook.org/ns/docbook'


def default_mendeley_database():
    '''
    Search for a default Mendeley sqlite database.
    '''
    # Linux
    dirs = [os.path.expanduser('~/.local/share/data/Mendeley Ltd./Mendeley Desktop')]
    # Windows Vista/7 (untested)
    if 'LOCALAPPDATA' in os.environ:
        dirs.append(os.path.join(os.environ('LOCALAPPDATA'),
                                 'Mendeley Ltd.', 'Mendeley Desktop'))
    # Windows XP (untested)
    dirs.append(os.path.join(os.path.expanduser('~'),
                             'Local Settings', 'Application Data',
                             'Mendeley Ltd.', 'Mendeley Desktop'))
    # MacOS X
    dirs.append(os.path.join(os.path.expanduser('~'),
                             'Library', 'Application Support',
                             'Mendeley Desktop'))
    
    for d in dirs:
      for path in pathlib.Path(d).glob('*@www.mendeley.com.sqlite'):
          return str(path)
    return None

def default_bibtex_database(doc):
    for d in [os.path.dirname(doc)]:
        for path in pathlib.Path(d).glob('*.bib'):
            return str(path)
    return None


def process_dom(dom, db_type='mendeley', db=None, db_encoding='utf-8',
                citation_separator=';', sort_order='alpha',
                style='harvard1',
                link_format=lambda x: x.lower()):
    '''
    mendeley, json and bibtex bibliography database are supported.
    
    :param xml.dom.Document dom:   Source DOM document.
    :param str db_type:            Bibliography database type.
    :param str db:                 Bibliography database.
    :param str db_encoding:        Bibliography database encoding.
    :param str citation_separator: Separator delimiting multiple citations
                                   in a single <citation/> element.
    :param str style:              CSL style to use.
    '''
    # Select a default bibliography database
    if db is None and db_type == 'mendeley':
        db = bibgen.default_mendeley_database()

    # Search all citations and their keys
    cit_nodes = [(n.childNodes[0].data.lower().split(citation_separator),n)
                 for n in dom.getElementsByTagNameNS(docbook_ns, 'citation') 
                 if len(n.childNodes) == 1
                    and n.childNodes[0].nodeType == n.TEXT_NODE]

    # Setup a citeproc context
    bib_style = cp.CitationStylesStyle(style)
    if db_type == 'mendeley':
        from bibgen.citeproc.mendeley import CiteProcMendeley
        bib_src = CiteProcMendeley(db)
    elif db_type == 'json':
        json_data = json.load(open(db))
        bib_src = citeproc.source.json.CiteProcJSON(json_data)
    # Default to BibTex
    else: 
        # Check for bibtexparser
        try:
            import bibtexparser
            with codecs.open(db, 'r', db_encoding) as fp:
                from bibgen.citeproc.bibtexparser import CiteProcBibTeX
                bib_src = CiteProcBibTeX(fp)
        # Use minimalistic bibtex parser from citeproc-py
        except ImportError:
            print('warning: defaulting to citeproc-py’s bibtex parser, you may want to install bibtexparser')
            with codecs.open(db, 'r', db_encoding) as fp:
                bib_src = citeproc.source.bibtex.BibTeX(fp)
    biblio = cp.CitationStylesBibliography(bib_style,
                                           bib_src, bibgen.formatter)

    # Create the citations
    cits = []
    for (keys,n) in cit_nodes:
        def mkitem(key):
            return cp.CitationItem(key,
                     prefix='<link linkend="%s">'%(link_format(key)),
                     suffix='</link>')
        cit = cp.Citation([mkitem(key) for key in keys])
        biblio.register(cit)
        cits.append(cit)

    # Print the citations
    def warn(cit_item):
        print('warning: citation reference not found for', cit_item.key)
    for ((keys,n),cit) in zip(cit_nodes,cits):
        txt = str(biblio.cite(cit,warn))
        cit_dom = xml.dom.minidom.parseString('<para>%s</para>'%txt)
        # Remove all previous <citation/> children
        for node in n.childNodes:
            n.removeChild(node)
        # Add the new children
        cit_root = cit_dom.documentElement
        while cit_root.hasChildNodes():
            n.appendChild(cit_root.firstChild)

    # Fill the existing bibliography node
    # TODO: add support for per-section bibliographies
    for bib_node in dom.getElementsByTagNameNS(docbook_ns, 'bibliography'):
        # Only keep the bibliography title
        # TODO: specify which nodes to keep
        for node in bib_node.childNodes:
            if node.nodeType == node.ELEMENT_NODE:
                if node.tagName != 'title':
                    bib_node.removeChild(node)
        # Add the bibliography entries
        bib_entries = list(zip(biblio.items,biblio.bibliography()))
        if sort_order == 'alpha':
            bib_entries.sort(key=lambda x: x[0].key)
        # TODO add support for raw entries
        for (itm,bibitem) in bib_entries:
            itemdom = xml.dom.minidom.parseString('<bibliomixed xml:id="%s">%s</bibliomixed>'%(link_format(itm.key),str(bibitem)))
            bib_node.appendChild(itemdom.documentElement)
        # Only process one bibliography element
        break



# DocBook formatter for citeproc-py

from citeproc.formatter.html import TagWrapper, preformat

class Italic(TagWrapper):
    tag = 'emphasis'

class Oblique(TagWrapper):
    tag = 'emphasis'
    attributes = {'role': 'oblique'}

class Bold(TagWrapper):
    tag = 'emphasis'
    attributes = {'role': 'bold'}

class Light(TagWrapper):
    pass

class Underline(TagWrapper):
    tag = 'emphasis'
    attributes = {'role': 'underline'}

class Superscript(TagWrapper):
    tag = 'superscript'

class Subscript(TagWrapper):
    tag = 'subscript'

class SmallCaps(TagWrapper):
    tag = 'emphasis'
    attributes = {'role': 'smallcaps'}



import sqlite3
import sys

import citeproc
import citeproc.source
import citeproc.types


types = {'Book': citeproc.types.BOOK,
         'BookSection': citeproc.types.CHAPTER,
         'ConferenceProceedings': citeproc.types.PAPER_CONFERENCE,
         'JournalArticle': citeproc.types.ARTICLE_JOURNAL,
         'Report': citeproc.types.REPORT,
         'Thesis': citeproc.types.THESIS}

def parse_type(name):
    if name in types:
        return types[name]
    return None

class CiteProcMendeley(citeproc.source.BibliographySource):
    def __init__(self, database):
        # Open the sqlite database
        conn = sqlite3.connect(database)
        conn.row_factory = sqlite3.Row
        curs = conn.cursor()

        # Fetch all documents
        curs.execute('SELECT id,citationKey,type,publication,year,month,volume,issue,pages,title FROM Documents')
        docs = curs.fetchall()
        for doc in docs:
            # Check for a citation key
            if doc['citationKey'] is None:
                continue

            # Check type
            type = parse_type(doc['type'])
            if type is None:
                print('warning: skipping unknown document type ‘%s’ for %s'%(doc['type'],doc['citationKey']),
                      file=sys.stderr)
                continue
            
            # Create a date
            if doc['year'] is None:
                print('warning: skipping undated publication for', doc['citationKey'], file=sys.stderr)
                continue
            date = {'year': int(doc['year'])}
            if doc['month'] is not None:
                date['month'] = doc['month']

            # Create document metadata
            meta = {'container_title': doc['publication'],
                    'issued': citeproc.source.Date(**date),
                    'title': doc['title'],
                    'author': []
                   }

            # Other metadata
            for key in ['issue', 'volume']:
                if doc[key] is not None:
                    meta[key] = doc[key]

            # Fetch authors
            curs.execute('SELECT firstNames,lastName FROM DocumentContributors WHERE documentId=? AND contribution="DocumentAuthor"', (doc['id'],))
            authors = curs.fetchall()
            for author in authors:
                name = citeproc.source.Name(family=author['lastName'], given=author['firstNames'])
                meta['author'].append(name)

            # Parse pages
            if doc['pages'] is not None:
                pp = doc['pages'].split('-')
                if len(pp) == 1:
                    meta['page'] = citeproc.source.Pages(first=pp[0])
                else:
                    meta['page'] = citeproc.source.Pages(first=pp[0], last=pp[1])

            # Register the document
            self.add(citeproc.source.Reference(doc['citationKey'].lower(), parse_type(doc['type']), **meta))



#! /usr/bin/env python3

from setuptools import setup

setup(name='bibgen',
      version='0.1.1',
      description='A citation and bibliography generator',
      long_description='''bibgen is a Python script that allows styled
      citations and bibliography generation using external
      bibliographic databases. It supports databases such as BibTeX,
      JSON or Mendeley, and generates styled citations and
      bibliography sections for several document types including
      DocBook. Textual citations and entries in the generated
      bibliographies are styled using rules found in CSL
      stylesheets. Thousands of those stylesheets are readily
      available, for instance through the Zotero style repository.''',
      author='Émilien Tlapale',
      author_email='emilien@tlapale.com',
      url='http://emilien.tlapale.com/code/bibgen',
      packages=['bibgen',
                'bibgen.citeproc'],
      scripts=['bin/bibgen'],
      install_requires = ['citeproc-py'],
      extras_require = {
        'enhanced_bibtex':  ['bibtexparser']
      },
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Documentation',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Text Processing',
          'Topic :: Text Processing :: Markup :: XML',
      ],
     )



import warnings

import citeproc.source
import citeproc.source.bibtex.bibtex as cbt
import citeproc.source.bibtex.bibparse as cbp

import bibtexparser
import bibtexparser.bparser
import bibtexparser.customization


class CiteProcBibTeX(cbt.BibTeX):
    '''
    BibTeX source for citeproc-py based on bibtexparser.
    '''
    def __init__(self, fp):
        parser = bibtexparser.bparser.BibTexParser()
        parser.customization = bibtexparser.customization.convert_to_unicode
        bib_db = bibtexparser.load(fp, parser=parser)
        for entry in bib_db.entries:
            bent = cbp.BibTeXEntry(entry['type'], entry)
            self.add(self.create_reference(entry['id'].lower(), bent))

    # This functions is copy/pasted/modified from citeproc-py
    def _bibtex_to_csl(self, bibtex_entry):
        csl_dict = {}
        for field, value in bibtex_entry.items():
            try:
                csl_field = self.fields[field]
            except KeyError:
                if field not in ('year', 'month', 'filename'):
                    warnings.warn("Unsupported BibTeX field '{}'".format(field))
                continue
            if field in ('number', 'volume'):
                try:
                    value = int(value)
                except ValueError:
                    pass
            elif field == 'pages':
                if value.endswith('+'):
                    value = citeproc.source.Pages(first=value[:-1])
                else:
                    fl = value.replace(' ', '').split('--')
                    if len(fl) > 1:
                        value = citeproc.source.Pages(first=fl[0], last=fl[1])
                    else:
                        value = citeproc.source.Pages(first=fl[0])
            elif field in ('author', 'editor'):
                value = [name for name in self._parse_author(value)]
            else:
                try:
                    value = self._parse_title(value)
                except TypeError:
                    value = str(value)
            csl_dict[csl_field] = value
        return csl_dict

